@@ -24,30 +24,56 @@
  * Revised by Ryusuke Konishi <ryusuke@osrg.net>.
  */
 
-#define _FILE_OFFSET_BITS 64
-//#define _LARGEFILE_SOURCE
-#define _LARGEFILE64_SOURCE
-#define __USE_FILE_OFFSET64
-#define _XOPEN_SOURCE 600
-
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif	/* HAVE_CONFIG_H */
 
 #undef CONFIG_ATIME_FILE
-#undef CONFIG_SKETCH_FILE
 
 #include <stdio.h>
+
+#if HAVE_STDLIB_H
 #include <stdlib.h>
+#endif	/* HAVE_STDLIB_H */
+
 #include <assert.h>
+
+#if HAVE_UNISTD_H
 #include <unistd.h>
+#endif	/* HAVE_UNISTD_H */
+
+#if HAVE_FCNTL_H
 #include <fcntl.h>
+#endif	/* HAVE_FCNTL_H */
+
+#if HAVE_STRINGS_H
 #include <strings.h>
+#endif	/* HAVE_STRINGS_H */
+
 #include <stdarg.h>
+
+#if HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif	/* HAVE_SYS_IOCTL_H */
+
+#if HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif	/* HAVE_SYS_STAT_H */
+
+#if HAVE_SYS_WAIT_H
 #include <sys/wait.h>
+#endif	/* HAVE_SYS_WAIT_H */
+
 #include <uuid/uuid.h>
+
+#if HAVE_STRING_H
 #include <string.h>
+#endif	/* HAVE_STRING_H */
+
 #include <errno.h>
+
 #include "mkfs.h"
+#include "nilfs_feature.h"
 
 
 typedef __u64  blocknr_t;
@@ -57,6 +83,9 @@ typedef __u64  blocknr_t;
 #define max_t(type,x,y) \
 	({ type __x = (x); type __y = (y); __x > __y ? __x : __y; })
 
+extern __u32 crc32_le(__u32 seed, unsigned char const *data, size_t length);
+#define nilfs_crc32(seed, data, length)  crc32_le(seed, data, length)
+
 /*
  * System primitives
  */
@@ -71,32 +100,33 @@ static const char badblocks[] = "/sbin/" BADBLOCKS_NAME;
 /*
  * Command interface primitives
  */
-#define _MI_  "\n       "   /* Message indent */
-
-struct mkfs_options {
-	long blocksize;
-	long blocks_per_segment;
-	long r_segments_percentage;
-	int quiet, cflag, nflag;
-	time_t ctime;
-	char volume_label[16];
-};
-
 extern char *optarg;
 extern int optind;
 
 char *progname = "mkfs.nilfs2";
 
-struct mkfs_options options = {
-	.blocksize = NILFS_DEF_BLOCKSIZE,
-	.blocks_per_segment = NILFS_DEF_BLKS_PER_SEG,
-	.r_segments_percentage = NILFS_DEF_RESERVED_SEGMENTS,
-	.quiet = 0,
-	.cflag = 0,
-	.nflag = 0,
+/* Options */
+static int quiet = 0;
+static int cflag = 0;
+static int nflag = 0;
+static int verbose = 0;
+static int discard = 1;
+static unsigned long blocksize = NILFS_DEF_BLOCKSIZE;
+static unsigned long blocks_per_segment = NILFS_DEF_BLKS_PER_SEG;
+static unsigned long r_segments_percentage = NILFS_DEF_RESERVED_SEGMENTS;
+
+static time_t creation_time = 0;
+static char volume_label[80] = {0};
+static __u64 compat_array[NILFS_MAX_FEATURE_TYPES] = {
+	/* Compat */
+	0,
+	/* Read-only compat */
+	0,
+	/* Incompat */
+	0
 };
 
-static void parse_options(int argc, char *argv[], struct mkfs_options *opts);
+static void parse_options(int argc, char *argv[]);
 
 
 /*
@@ -129,8 +159,7 @@ static void parse_options(int argc, char *argv[], struct mkfs_options *opts);
 static const unsigned group_desc_blocks_per_group = 1;
 static const unsigned bitmap_blocks_per_group = 1;
 static const unsigned nr_initial_segments = 2; /* initial segment + next */
-static const unsigned nr_initial_inodes = 3;  /* root directory + .sketch +
-						 .nilfs */
+static const unsigned nr_initial_inodes = 2;  /* root directory + .nilfs */
 static const __u64 first_cno = 1; /* Number of the first checkpoint */
 
 /* Segment layout information (per partial segment) */
@@ -153,13 +182,10 @@ struct nilfs_segment_info {
 	struct nilfs_file_info files[MAX_FILES];
 
 	unsigned        nvblocknrs;
-	unsigned        nblocks_used;
 	/* + super root (1 block)*/
 };
 
 /* Disk layout information */
-static long blocksize;
-
 struct nilfs_disk_info {
 	const char      *device;
 	__u64           dev_size;
@@ -188,9 +214,8 @@ struct nilfs_segment_ref {
 	__u64 cno;                /* checkpoint number */
 };
 
-static void init_disk_layout(struct nilfs_disk_info *, int, const char *,
-			     struct mkfs_options *);
-
+static void init_disk_layout(struct nilfs_disk_info *di, int fd,
+			     const char *device);
 
 static inline blocknr_t count_free_blocks(struct nilfs_disk_info *di)
 {
@@ -211,22 +236,21 @@ segment_start_blocknr(struct nilfs_disk_info *di, unsigned long segnum)
 static void **disk_buffer = NULL;
 static unsigned long disk_buffer_size;
 
-static void init_disk_buffer(long);
+static void init_disk_buffer(long max_blocks);
 static void destroy_disk_buffer(void);
-static void *map_disk_buffer(blocknr_t, int);
+static void *map_disk_buffer(blocknr_t blocknr, int clear_flag);
 
-static void read_disk_header(int, const char *);
-static void write_disk(int, struct nilfs_disk_info *, struct mkfs_options *);
+static void read_disk_header(int fd, const char *device);
+static void write_disk(int fd, struct nilfs_disk_info *di);
 
 /*
  * Routines to format blocks
  */
 static struct nilfs_super_block *raw_sb;
 
-static void prepare_super_block(struct nilfs_disk_info *,
-				struct mkfs_options *);
-static void commit_super_block(struct nilfs_disk_info *,
-			       const struct nilfs_segment_ref *);
+static void prepare_super_block(struct nilfs_disk_info *di);
+static void commit_super_block(struct nilfs_disk_info *di,
+			       const struct nilfs_segment_ref *segref);
 
 struct nilfs_fs_info {
 	struct nilfs_disk_info *diskinfo;
@@ -247,13 +271,12 @@ struct nilfs_fs_info {
 
 static struct nilfs_fs_info nilfs;
 
-static void init_nilfs(struct nilfs_disk_info *);
+static void init_nilfs(struct nilfs_disk_info *di);
 static void prepare_segment(struct nilfs_segment_info *);
 static void commit_segment(void);
-static void make_rootdir(void);
-static void make_sketch(void);
-static void make_dot_nilfs(void);
-static void make_reserved_files(void);
+static void nilfs_mkfs_make_rootdir(void);
+static void nilfs_mkfs_make_dot_nilfs(void);
+static void nilfs_mkfs_make_reserved_files(void);
 
 
 static inline struct nilfs_segment_ref *get_last_segment(void)
@@ -272,16 +295,64 @@ static void cannot_allocate_memory(void);
 static void too_small_segment(unsigned long, unsigned long);
 
 /* I/O routines */
-static void disk_scan(const char *, struct mkfs_options *);
-static void check_mount(int, const char *);
+#ifdef __linux__
+
+#ifndef BLKDISCARD
+#define BLKDISCARD	_IO(0x12,119)
+#endif
+
+#ifndef BLKDISCARDZEROES
+#define BLKDISCARDZEROES _IO(0x12,124)
+#endif
+
+/**
+ * nilfs_mkfs_discard_range - issue discard command to the device
+ * @fd: file descriptor of the device
+ * @start: start offset of the region to discard (in bytes)
+ * @len: length of the region to discard (in bytes)
+ *
+ * Returns zero if the discard succeeds.  Otherwise, -1 is returned.
+ */
+static int nilfs_mkfs_discard_range(int fd, __u64 start, __u64 len)
+{
+	__u64 range[2] = { start, len };
+	int ret;
+
+	ret = ioctl(fd, BLKDISCARD, &range);
+	if (verbose) {
+		pinfo("Discard device from %llu to %llu: %s.",
+		      (unsigned long long)start,
+		      (unsigned long long)start + len,
+		      ret ? "failed" : "succeeded");
+	}
+	return ret;
+}
+
+/**
+ * nilfs_mkfs_discard_zeroes_data - get if discarded blocks are zeroed or not
+ * @fd: file descriptor of the device
+ */
+static int nilfs_mkfs_discard_zeroes_data(int fd)
+{
+	int discard_zeroes_data = 0;
+
+	ioctl(fd, BLKDISCARDZEROES, &discard_zeroes_data);
+	return discard_zeroes_data;
+}
+#else
+#define nilfs_mkfs_discard_range(fd, start, len)	1
+#define nilfs_mkfs_discard_zeroes_data(fd)		0
+#endif
+
+static void disk_scan(const char *device);
+static void check_mount(int fd, const char *device);
 
 
 /*
  * Routines to decide disk layout
  */
-static unsigned
-count_blockgrouped_file_blocks(long blocksize, unsigned entry_size,
-			       unsigned nr_initial_entries)
+static unsigned count_blockgrouped_file_blocks(unsigned entry_size,
+					       unsigned nr_initial_entries)
 {
 	unsigned long entries_per_block = blocksize / entry_size;
 
@@ -289,13 +360,12 @@ count_blockgrouped_file_blocks(long blocksize, unsigned entry_size,
 		ROUNDUP_DIV(nr_initial_entries, entries_per_block);
 }
 
-static unsigned count_ifile_blocks(long blocksize)
+static unsigned count_ifile_blocks(void)
 {
 	unsigned long entries_per_group = blocksize * 8; /* CHAR_BIT */
 	unsigned nblocks;
 
-	nblocks = count_blockgrouped_file_blocks(blocksize,
-						 sizeof(struct nilfs_inode),
+	nblocks = count_blockgrouped_file_blocks(sizeof(struct nilfs_inode),
 						 NILFS_MAX_INITIAL_INO);
 	if (NILFS_MAX_INITIAL_INO > entries_per_group ||
 	    nblocks > NILFS_MAX_BMAP_ROOT_PTRS)
@@ -303,7 +373,7 @@ static unsigned count_ifile_blocks(long blocksize)
 	return nblocks;
 }
 
-static unsigned count_sufile_blocks(long blocksize)
+static unsigned count_sufile_blocks(void)
 {
 	unsigned long sufile_segment_usages_per_block
 		= blocksize / sizeof(struct nilfs_segment_usage);
@@ -312,7 +382,7 @@ static unsigned count_sufile_blocks(long blocksize)
 			   sufile_segment_usages_per_block);
 }
 
-static unsigned count_cpfile_blocks(long blocksize)
+static unsigned count_cpfile_blocks(void)
 {
 	const unsigned nr_initial_checkpoints = 1;
 	unsigned long cpfile_checkpoints_per_block
@@ -323,13 +393,13 @@ static unsigned count_cpfile_blocks(long blocksize)
 			   cpfile_checkpoints_per_block);
 }
 
-static unsigned count_dat_blocks(long blocksize, unsigned nr_dat_entries)
+static unsigned count_dat_blocks(unsigned nr_dat_entries)
 {
 	unsigned long entries_per_group = blocksize * 8; /* CHAR_BIT */
 	unsigned nblocks;
 
 	nblocks = count_blockgrouped_file_blocks(
-		blocksize, sizeof(struct nilfs_dat_entry), nr_dat_entries);
+		sizeof(struct nilfs_dat_entry), nr_dat_entries);
 	if (nr_dat_entries > entries_per_group ||
 	    nblocks > NILFS_MAX_BMAP_ROOT_PTRS)
 		perr("Internal error: too many initial dat entries");
@@ -349,8 +419,8 @@ static void nilfs_check_ondisk_sizes(void)
 }
 
 static unsigned long
-__increment_segsum_size(unsigned long offset, long blocksize, 
-			unsigned item_size, unsigned count)
+__increment_segsum_size(unsigned long offset, unsigned item_size,
+			unsigned count)
 {
 	unsigned long offset2;
 	unsigned rest_items_in_block =
@@ -369,16 +439,15 @@ __increment_segsum_size(unsigned long offset, long blocksize,
 	return offset2;
 }
 
-static void
-increment_segsum_size(struct nilfs_segment_info *si, long blocksize,
-		      unsigned nblocks_in_file, int dat_flag)
+static void increment_segsum_size(struct nilfs_segment_info *si,
+				  unsigned nblocks_in_file, int dat_flag)
 {
 	unsigned binfo_size = dat_flag ? 
 		sizeof(__le64) /* offset */ : sizeof(struct nilfs_binfo_v);
 
-	si->sumbytes = __increment_segsum_size(si->sumbytes, blocksize,
+	si->sumbytes = __increment_segsum_size(si->sumbytes,
 					       sizeof(struct nilfs_finfo), 1);
-	si->sumbytes = __increment_segsum_size(si->sumbytes, blocksize,
+	si->sumbytes = __increment_segsum_size(si->sumbytes,
 					       binfo_size, nblocks_in_file);
 }
 
@@ -398,9 +467,8 @@ static unsigned long nilfs_min_nsegments(struct nilfs_disk_info *di, long rp)
 		max_t(unsigned long, nr_initial_segments, NILFS_MIN_NUSERSEGS);
 }
 
-static void
-init_disk_layout(struct nilfs_disk_info *di, int fd, const char *device,
-		 struct mkfs_options *opts)
+static void init_disk_layout(struct nilfs_disk_info *di, int fd,
+			     const char *device)
 {
 	__u64 dev_size;
 	time_t nilfs_time = time(NULL);
@@ -420,14 +488,14 @@ init_disk_layout(struct nilfs_disk_info *di, int fd, const char *device,
 
 	di->device = device;
 	di->dev_size = dev_size;
-	di->blkbits = my_log2(opts->blocksize);
-	di->ctime = (opts->ctime ? : nilfs_time);
+	di->blkbits = my_log2(blocksize);
+	di->ctime = (creation_time ? : nilfs_time);
 	srand48(nilfs_time);
 	di->crc_seed = (__u32)mrand48();
 
-	di->blocks_per_segment = opts->blocks_per_segment;
-	segment_size = di->blocks_per_segment * opts->blocksize;
-	first_segblk = ROUNDUP_DIV(NILFS_DISKHDR_SIZE, opts->blocksize);
+	di->blocks_per_segment = blocks_per_segment;
+	segment_size = di->blocks_per_segment * blocksize;
+	first_segblk = ROUNDUP_DIV(NILFS_DISKHDR_SIZE, blocksize);
 	di->first_segment_block = first_segblk;
 	if (first_segblk + NILFS_PSEG_MIN_BLOCKS > di->blocks_per_segment)
 		too_small_segment(di->blocks_per_segment,
@@ -435,12 +503,12 @@ init_disk_layout(struct nilfs_disk_info *di, int fd, const char *device,
 
 	di->nsegments = (NILFS_SB2_OFFSET_BYTES(dev_size) >> di->blkbits) /
 		di->blocks_per_segment;
-	min_nsegments = nilfs_min_nsegments(di, opts->r_segments_percentage);
+	min_nsegments = nilfs_min_nsegments(di, r_segments_percentage);
 	if (di->nsegments < min_nsegments)
-		perr("Error: too small device."
-		     _MI_ "device size=%llu bytes, required size=%llu bytes."
-		     _MI_ "Please enlarge the device, "
-		          "or shorten segments with -B option.", dev_size,
+		perr("Error: too small device.\n"
+		     "       device size=%llu bytes, required size=%llu bytes.\n"
+		     "       Please enlarge the device, "
+		     "or shorten segments with -B option.", dev_size,
 		     (unsigned long long)segment_size * min_nsegments);
 	di->nseginfo = 0;
 
@@ -472,7 +540,6 @@ static void fix_disk_layout(struct nilfs_disk_info *di)
 		struct nilfs_segment_info *si = &di->seginfo[i];
 		blocknr_t blocknr = si->start + si->nblk_sum;
 
-		si->nblocks_used = (di->nblocks_used += si->nblocks);
 		si->nblocks += si->nblk_sum + 1 /* summary and super root */;
 		if (si->nblocks > di->blocks_per_segment)
 			too_small_segment(di->blocks_per_segment, si->nblocks);
@@ -484,6 +551,11 @@ static void fix_disk_layout(struct nilfs_disk_info *di)
 				continue;
 			fi->start = blocknr;
 			blocknr += fi->nblocks;
+
+			if (fi->ino != NILFS_DAT_INO &&
+			    fi->ino != NILFS_SUFILE_INO &&
+			    fi->ino != NILFS_CPFILE_INO)
+				di->nblocks_used += fi->nblocks;
 		}
 		if (di->nblocks_to_write < si->start + si->nblocks)
 			di->nblocks_to_write = si->start + si->nblocks;
@@ -509,7 +581,7 @@ static void add_file(struct nilfs_segment_info *si, ino_t ino,
 	si->nblocks += nblocks;
 	if (nblocks > 0) {
 		si->nfinfo++;
-		increment_segsum_size(si, blocksize, nblocks, dat_flag);
+		increment_segsum_size(si, nblocks, dat_flag);
 		if (!dat_flag)
 			si->nvblocknrs += nblocks;
 		si->nblk_sum = ROUNDUP_DIV(si->sumbytes, blocksize);
@@ -520,43 +592,39 @@ static void add_file(struct nilfs_segment_info *si, ino_t ino,
 int main(int argc, char *argv[])
 {
 	struct nilfs_disk_info diskinfo, *di = &diskinfo;
-	struct mkfs_options *opts = &options;
 	struct nilfs_segment_info *si;
 	struct stat statbuf;
 	const char *device;
 	int fd;
 
-	parse_options(argc, argv, opts);
+	parse_options(argc, argv);
 	device = argv[optind];
 
-	blocksize = opts->blocksize;
-
 	if (stat(device, &statbuf) != 0)
 		perr("Error: cannot find %s: %s", device, strerror(errno));
 	else if (!S_ISREG(statbuf.st_mode) && !S_ISBLK(statbuf.st_mode))
 		perr("Error: device must be a block device or a file");
 
-	if (opts->cflag)
-		disk_scan(device, opts);  /* check the block device */
+	if (cflag)
+		disk_scan(device);  /* check the block device */
 	if ((fd = open(device, O_RDWR)) < 0)
 		perr("Error: cannot open device: %s", device);
 	check_mount(fd, device);
 
-	init_disk_layout(di, fd, device, opts);
+	init_disk_layout(di, fd, device);
 	si = new_segment(di);
 
 	add_file(si, NILFS_ROOT_INO, 1, 0);
-	add_file(si, NILFS_SKETCH_INO, 0, 0);
 	add_file(si, NILFS_NILFS_INO, 0, 0);
 	add_file(si, NILFS_ATIME_INO, 0, 0);
 	add_file(si, 1, 0, 0);
 	add_file(si, 8, 0, 0);
 	add_file(si, 9, 0, 0);
-	add_file(si, NILFS_IFILE_INO, count_ifile_blocks(blocksize), 0);
-	add_file(si, NILFS_CPFILE_INO, count_cpfile_blocks(blocksize), 0);
-	add_file(si, NILFS_SUFILE_INO, count_sufile_blocks(blocksize), 0);
-	add_file(si, NILFS_DAT_INO,
-		 count_dat_blocks(blocksize, si->nvblocknrs), 1);
+	add_file(si, 10, 0, 0);
+	add_file(si, NILFS_IFILE_INO, count_ifile_blocks(), 0);
+	add_file(si, NILFS_CPFILE_INO, count_cpfile_blocks(), 0);
+	add_file(si, NILFS_SUFILE_INO, count_sufile_blocks(), 0);
+	add_file(si, NILFS_DAT_INO, count_dat_blocks(si->nvblocknrs), 1);
 
 	fix_disk_layout(di);
 
@@ -564,19 +632,18 @@ int main(int argc, char *argv[])
 	init_disk_buffer(di->nblocks_to_write);
 	read_disk_header(fd, device);
 
-	prepare_super_block(di, opts);
+	prepare_super_block(di);
 	init_nilfs(di);
 
 	prepare_segment(&di->seginfo[0]);
-	make_sketch();   /* Make sketch file */
-	make_dot_nilfs(); /* Make .nilfs */
-	make_rootdir();  /* Make root directory */
-	make_reserved_files();
+	nilfs_mkfs_make_dot_nilfs(); /* Make .nilfs */
+	nilfs_mkfs_make_rootdir();  /* Make root directory */
+	nilfs_mkfs_make_reserved_files();
 	commit_segment();
 
 	commit_super_block(di, get_last_segment());
 
-	write_disk(fd, di, opts); /* Writing to the device */
+	write_disk(fd, di); /* Writing to the device */
 
 	close(fd);
 	exit(0);
@@ -585,7 +652,7 @@ int main(int argc, char *argv[])
 /*
  * I/O routines & primitives
  */
-static void disk_scan(const char *device, struct mkfs_options *opts)
+static void disk_scan(const char *device)
 {
 	struct stat statbuf;
 	pid_t pid;
@@ -596,7 +663,7 @@ static void disk_scan(const char *device, struct mkfs_options *opts)
 		return;
 	}
 
-	if (!opts->quiet)
+	if (!quiet)
 		pinfo("checking blocks");
 
 	pid = fork();
@@ -612,15 +679,14 @@ static void disk_scan(const char *device, struct mkfs_options *opts)
 			perr("Error: failed to drop setuid privileges");
 		args[i++] = badblocks;
 		args[i++] = "-b";
-		n = snprintf(bszbuf, BLOCKSIZE_BUFFER_SIZE, "%ld",
-			     opts->blocksize);
+		n = snprintf(bszbuf, BLOCKSIZE_BUFFER_SIZE, "%ld", blocksize);
 		if (n < 0 || n >= BLOCKSIZE_BUFFER_SIZE)
 			perr("Internal error: blocksize buffer overflow");
 		args[i++] = bszbuf;
 
-		if (!opts->quiet)
+		if (!quiet)
 			args[i++] = "-s";
-		if (opts->cflag > 1)
+		if (cflag > 1)
 			args[i++] = "-w";
 		args[i++] = device;
 		args[i] = NULL;
@@ -709,7 +775,7 @@ static void read_disk_header(int fd, const char *device)
 {
 	int i, hdr_blocks = ROUNDUP_DIV(NILFS_SB_OFFSET_BYTES, blocksize);
 
-	lseek64(fd, 0, SEEK_SET);
+	lseek(fd, 0, SEEK_SET);
 	for (i = 0; i < hdr_blocks; i++) {
 		if (read(fd, map_disk_buffer(i, 0), blocksize) < 0)
 			cannot_rw_device(fd, device, 1);
@@ -718,7 +784,7 @@ static void read_disk_header(int fd, const char *device)
 
 #define MAX_NBLOCKS_CLEAR_BUFFER	8
 
-static int erase_disk_range(int fd, off64_t offset, size_t count)
+static int erase_disk_range(int fd, off_t offset, size_t count)
 {
 	void *buffer;
 	size_t size, bufsz;
@@ -735,7 +801,7 @@ static int erase_disk_range(int fd, off64_t offset, size_t count)
 
 	memset(buffer, 0, bufsz);
 
-	if (lseek64(fd, offset, SEEK_SET) < 0)
+	if (lseek(fd, offset, SEEK_SET) < 0)
 		goto failed;
 
 	while (count > 0) {
@@ -753,43 +819,61 @@ static int erase_disk_range(int fd, off64_t offset, size_t count)
 
 static int erase_disk(int fd, struct nilfs_disk_info *di)
 {
+	const unsigned int sector_size = 512;
+	off_t start, end;
 	int ret;
 
-	BUG_ON(di->dev_size < NILFS_DISK_ERASE_SIZE ||
-	       di->dev_size - NILFS_DISK_ERASE_SIZE < NILFS_SB_OFFSET_BYTES);
+	/*
+	 * Define range of the partition that nilfs uses.  This should
+	 * not depend on the type of underlying device.
+	 */
+	start = NILFS_SB_OFFSET_BYTES;
+	end = di->dev_size & ~((__u64)sector_size - 1);
+
+	BUG_ON(end < NILFS_DISK_ERASE_SIZE ||
+	       end - NILFS_DISK_ERASE_SIZE < start);
+
+	if (discard) {
+		ret = nilfs_mkfs_discard_range(fd, start, end - start);
+		if (!ret && nilfs_mkfs_discard_zeroes_data(fd)) {
+			if (verbose)
+				pinfo("Discard succeeded and will return 0s "
+				      " - skip wiping");
+			goto out;
+		}
+	}
 
 	/* Erase tail of partition */
-	ret = erase_disk_range(fd, di->dev_size - NILFS_DISK_ERASE_SIZE,
+	ret = erase_disk_range(fd, end - NILFS_DISK_ERASE_SIZE,
 			       NILFS_DISK_ERASE_SIZE);
 	if (ret == 0) {
 		/* Erase head of partition */
-		ret = erase_disk_range(fd, NILFS_SB_OFFSET_BYTES,
-				       NILFS_DISK_ERASE_SIZE -
-				       NILFS_SB_OFFSET_BYTES);
+		ret = erase_disk_range(fd, start,
+				       NILFS_DISK_ERASE_SIZE - start);
 	}
+out:
 	return ret;
 }
 
-static void write_disk(int fd, struct nilfs_disk_info *di,
-		       struct mkfs_options *opts)
+static void write_disk(int fd, struct nilfs_disk_info *di)
 {
 	blocknr_t blocknr;
 	struct nilfs_segment_info *si;
 	int i;
 
-	if (!opts->quiet) {
+	if (!quiet) {
 		show_version();
-		pinfo("Start writing file system initial data to the device"
-		      _MI_ "Blocksize:%d  Device:%s  Device Size:%llu",
+		pinfo("Start writing file system initial data to the device\n"
+		      "       Blocksize:%d  Device:%s  Device Size:%llu",
 		      blocksize, di->device, di->dev_size);
 	}
-	if (!opts->nflag) {
+	if (!nflag) {
 		if (erase_disk(fd, di) < 0)
 			goto failed_to_write;
 
 		/* Writing segments */
 		for (i = 0, si = di->seginfo; i < di->nseginfo; i++, si++) {
-			lseek64(fd, si->start * blocksize, SEEK_SET);
+			lseek(fd, si->start * blocksize, SEEK_SET);
 			for (blocknr = si->start;
 			     blocknr < si->start + si->nblocks; blocknr++) {
 				if (write(fd, map_disk_buffer(blocknr, 1),
@@ -801,20 +885,20 @@ static void write_disk(int fd, struct nilfs_disk_info *di,
 			goto failed_to_write;
 
 		/* Writing primary super block */
-		if (lseek64(fd, NILFS_SB_OFFSET_BYTES, SEEK_SET) < 0 ||
+		if (lseek(fd, NILFS_SB_OFFSET_BYTES, SEEK_SET) < 0 ||
 		    write(fd, raw_sb, sizeof(*raw_sb)) < 0)
 			goto failed_to_write;
 
 		/* Writing secondary super block */
-		if (lseek64(fd, NILFS_SB2_OFFSET_BYTES(di->dev_size),
-			    SEEK_SET) < 0 ||
+		if (lseek(fd, NILFS_SB2_OFFSET_BYTES(di->dev_size),
+			  SEEK_SET) < 0 ||
 		    write(fd, raw_sb, sizeof(*raw_sb)) < 0)
 			goto failed_to_write;
 
 		if (fsync(fd) < 0)
 			goto failed_to_write;
 	}
-	if (!opts->quiet)
+	if (!quiet)
 		pinfo("File system initialization succeeded !! ");
 	return;
 
@@ -854,7 +938,7 @@ check_reserved_segments_percentage(long r_segments_percentage)
 		     r_segments_percentage);
 }
 
-static inline void check_ctime(long ctime)
+static inline void check_ctime(time_t ctime)
 {
 	if ((long)time(NULL) - (long)ctime < 0) {
 		char cbuf[26], *cbufp;
@@ -867,46 +951,75 @@ static inline void check_ctime(long ctime)
 	}
 }
 
-static void parse_options(int argc, char *argv[], struct mkfs_options *opts)
+const static __u64 ok_features[NILFS_MAX_FEATURE_TYPES] = {
+	/* Compat */
+	0,
+	/* Read-only compat */
+	NILFS_FEATURE_COMPAT_RO_BLOCK_COUNT,
+	/* Incompat */
+	0
+};
+
+static void nilfs_mkfs_edit_feature(const char *str)
+{
+	if (nilfs_edit_feature(str, compat_array, ok_features, NULL, NULL,
+			       NULL) < 0)
+		perr("Error: invalid feature set option: %s", str);
+}
+
+static void parse_options(int argc, char *argv[])
 {
 	int c, show_version_only = 0;
+	char *fs_features = NULL;
 
-	while ((c = getopt(argc, argv, "b:B:cL:m:nqVP:")) != EOF) {
+	while ((c = getopt(argc, argv, "b:B:cKL:m:nqvO:P:V")) != EOF) {
 		switch (c) {
 		case 'b':
-			opts->blocksize = atol(optarg);
-			check_blocksize(opts->blocksize);
+			blocksize = atol(optarg);
+			check_blocksize(blocksize);
 			break;
 		case 'B':
-			opts->blocks_per_segment = atol(optarg);
+			blocks_per_segment = atol(optarg);
 			break;
 		case 'c':
-			opts->cflag++;
+			cflag++;
+			break;
+		case 'K':
+			discard = 0;
 			break;
 		case 'L':
-			strncpy(opts->volume_label, optarg,
-				sizeof(opts->volume_label));
+			strncpy(volume_label, optarg, sizeof(volume_label));
 			break;
 		case 'm':
-			opts->r_segments_percentage = atol(optarg);
+			r_segments_percentage = atol(optarg);
 			break;
 		case 'n':
-			opts->nflag++;
+			nflag = 1;
 			break;
 		case 'q':
-			opts->quiet++;
+			quiet = 1;
 			break;
-		case 'V':
-			show_version_only++;
+		case 'v':
+			verbose++;
+			break;
+		case 'O':
+			fs_features = optarg;
 			break;
 		case 'P': /* Passive mode */
-			opts->ctime = atol(optarg);
-			check_ctime(opts->ctime);
+			creation_time = atol(optarg);
+			check_ctime(creation_time);
+			break;
+		case 'V':
+			show_version_only = 1;
 			break;
 		default:
 			usage();
 		}
 	}
+
+	if (quiet)
+		verbose = 0;
+
 	if ((optind == argc) && !show_version_only)
 		usage();
 
@@ -915,8 +1028,11 @@ static void parse_options(int argc, char *argv[], struct mkfs_options *opts)
 		exit(0);
 	}
 
-	check_blocks_per_segment(opts->blocks_per_segment);
-	check_reserved_segments_percentage(opts->r_segments_percentage);
+	check_blocks_per_segment(blocks_per_segment);
+	check_reserved_segments_percentage(r_segments_percentage);
+
+	if (fs_features)
+		nilfs_mkfs_edit_feature(fs_features);
 
         if (argc > 0) {
                 char *cp = strrchr(argv[0], '/');
@@ -932,16 +1048,16 @@ static void usage(void)
 {
 	fprintf(stderr,
 		"Usage: %s [-b block-size] [-B blocks-per-segment] [-c] \n"
-		"[-L volume-label] [-q] [-r revision-level]\n"
-		"[-m reserved-segments-percentage] [-V] device\n",
+		"[-L volume-label] [-m reserved-segments-percentage] \n"
+		"[-O feature[,...]] \n"
+		"[-nqvKV] device\n",
 		progname);
 	exit(1);
 }
 
 static void show_version(void)
 {
-	fprintf(stderr, "%s ver %d.%d\n", progname, NILFS_CURRENT_REV,
-		NILFS_MINOR_REV);
+	fprintf(stderr, "%s (%s %s)\n", progname, PACKAGE, PACKAGE_VERSION);
 }
 
 static void pinfo(const char *fmt, ...)
@@ -980,9 +1096,9 @@ static void cannot_allocate_memory(void)
 static void too_small_segment(unsigned long blocks_per_segment, 
 			      unsigned long required_blocks)
 {
-	perr("Error: too small segment."
-	     _MI_ "segment size=%lu blocks, required segment size=%lu blocks."
-	     _MI_ "Please enlarge segment with -B option.",
+	perr("Error: too small segment.\n"
+	     "       segment size=%lu blocks, required segment size=%lu blocks.\n"
+	     "       Please enlarge segment with -B option.",
 	     blocks_per_segment, required_blocks);
 }
 
@@ -1029,135 +1145,59 @@ static void inc_link_count(int ino)
 		= cpu_to_le16(le16_to_cpu(raw_inode->i_links_count) + 1);
 }
 
-static struct nilfs_dir_entry *next_dir_entry(struct nilfs_dir_entry *de)
+static volatile struct nilfs_dir_entry *
+next_dir_entry(volatile struct nilfs_dir_entry *de)
 {
-	return (void *)de + le16_to_cpu(de->rec_len);
+	return (void *)de + nilfs_rec_len_from_disk(de->rec_len);
 }
 
-#if 0 /* prepared for future use */
-static void
-add_dir_entry(ino_t dir_ino, ino_t ino, const char *name)
+static void nilfs_mkfs_make_rootdir(void)
 {
-	struct nilfs_file_info *fi = nilfs.files[dir_ino];
-	unsigned namelen = strlen(name);
-	unsigned reclen = NILFS_DIR_REC_LEN(namelen);
-	unsigned name_len, rec_len;
-	struct nilfs_dir_entry *de;
-	void *dir_end;
-	struct nilfs_inode *raw_inode = nilfs.files[ino]->raw_inode;
-	int type;
+	void *dirbuf = map_disk_buffer(nilfs.files[NILFS_ROOT_INO]->start, 1);
+	volatile struct nilfs_dir_entry *de = dirbuf;
+		/* volatile keyword is inserted to prevent failure of
+		   substitution to de->inode on a certain environment. */
+	unsigned rec_len, rec_end;
 
-	if (NILFS_NAME_LEN < namelen)
-		perr("Internal error: too long directory entry name: %s",
-		     name);
+	init_inode(NILFS_ROOT_INO, DT_DIR, 0755, blocksize);
 
-	de = map_disk_buffer(fi->start, 1);
-	dir_end = (void *)de + blocksize;
-
-	/* search the last entry */
-	while ((void *)de <= dir_end - reclen) {
-		name_len = NILFS_DIR_REC_LEN(de->name_len);
-		rec_len = le16_to_cpu(de->rec_len);
-		if (!de->inode && rec_len >= reclen)
-			goto got_it;
-		if (rec_len >= name_len + reclen)
-			goto got_it;
-		de = next_dir_entry(de);
-	} 
-	perr("Internal error: too many directory entry");
-	return;
- got_it:
-	if (de->inode) {
-		struct nilfs_dir_entry *de2 = (void *)de + name_len;
-		de2->rec_len = cpu_to_le16(rec_len - name_len);
-		de->rec_len = cpu_to_le16(name_len);
-		de = de2;
-	}
-	de->inode = cpu_to_le64(ino);
-	de->name_len = namelen & 0xff;
-	memcpy(de->name, name, NILFS_NAME_LEN);
-	
-	type = (le16_to_cpu(raw_inode->i_mode) >> 12) & 0xf;
-	switch (type) {
-	case DT_DIR:
-		de->file_type = NILFS_FT_DIR;
-		inc_link_count(dir_ino);
-		break;
-	case DT_REG:
-		de->file_type = NILFS_FT_REG_FILE;
-		break;
-	default:
-		perr("Internal error: file type (%d) not supported", type);
-	}
-}
-#endif
-
-static void make_empty_dir(ino_t dir_ino, ino_t parent_ino)
-{
-	struct nilfs_file_info *fi = nilfs.files[dir_ino];
-	struct nilfs_dir_entry *de = map_disk_buffer(fi->start, 1);
-	unsigned rec_len, rec_len2;
-
-	de->inode = cpu_to_le64(dir_ino);
+	de->inode = cpu_to_le64(NILFS_ROOT_INO);
 	de->name_len = 1;
-	rec_len2 = rec_len = NILFS_DIR_REC_LEN(1);
-	de->rec_len = cpu_to_le16(rec_len);
+	rec_end = rec_len = NILFS_DIR_REC_LEN(1);
+	de->rec_len = nilfs_rec_len_to_disk(rec_len);
 	de->file_type = NILFS_FT_DIR;
-	memcpy(de->name, ".\0\0\0\0\0\0", 8);
+	memcpy((void *)de->name, ".\0\0\0\0\0\0", 8);
 
 	de = next_dir_entry(de);
-	de->inode = cpu_to_le64(parent_ino);
+	de->inode = cpu_to_le64(NILFS_ROOT_INO);
 	de->name_len = 2;
-	rec_len2 += (rec_len = NILFS_DIR_REC_LEN(2));
-	de->rec_len = cpu_to_le16(rec_len);
+	rec_end += (rec_len = NILFS_DIR_REC_LEN(2));
+	de->rec_len = nilfs_rec_len_to_disk(rec_len);
 	de->file_type = NILFS_FT_DIR;
-	memcpy(de->name, "..\0\0\0\0\0", 8);
-//	rec_len += de->rec_len;
-
-#if defined(CONFIG_SKETCH_FILE)
-	de = next_dir_entry(de);
-	de->inode = cpu_to_le64(NILFS_SKETCH_INO);
-	de->name_len = 7;
-	rec_len2 += (rec_len = NILFS_DIR_REC_LEN(7));
-	de->rec_len = cpu_to_le16(rec_len);
-	de->file_type = NILFS_FT_REG_FILE;
-	memcpy(de->name, ".sketch", 8);
-//	rec_len += de->rec_len;
-#endif
+	memcpy((void *)de->name, "..\0\0\0\0\0", 8);
 
 	de = next_dir_entry(de);
 	de->inode = cpu_to_le64(NILFS_NILFS_INO);
 	de->name_len = 6;
-	de->rec_len = cpu_to_le16(blocksize - rec_len2);
+	de->rec_len = nilfs_rec_len_to_disk(blocksize - rec_end);
 	de->file_type = NILFS_FT_REG_FILE;
-	memcpy(de->name, ".nilfs\0", 8);
+	memcpy((void *)de->name, ".nilfs\0", 8);
+
+	inc_link_count(NILFS_ROOT_INO);
 }
 
-static void make_rootdir(void)
-{
-	const ino_t ino = NILFS_ROOT_INO;
-
-	init_inode(ino, DT_DIR, 0755, blocksize);
-	make_empty_dir(ino, ino);
-	inc_link_count(ino);
-}
-
-static void make_sketch(void)
-{
-	init_inode(NILFS_SKETCH_INO, DT_REG, 0644, 0);
-}
-
-static void make_dot_nilfs(void)
+static void nilfs_mkfs_make_dot_nilfs(void)
 {
 	init_inode(NILFS_NILFS_INO, DT_REG, 0644, 0);
 }
 
-static void make_reserved_files(void)
+static void nilfs_mkfs_make_reserved_files(void)
 {
 	init_inode(NILFS_ATIME_INO, DT_REG, 0, 0);
 	init_inode(1, DT_REG, 0, 0);
 	init_inode(8, DT_REG, 0, 0);
 	init_inode(9, DT_REG, 0, 0);
+	init_inode(10, DT_REG, 0, 0);
 }
 
 static void *
@@ -1476,6 +1516,7 @@ static void prepare_segment(struct nilfs_segment_info *si)
 	nilfs.segsum->ss_nblocks = cpu_to_le32(si->nblocks);
 	nilfs.segsum->ss_nfinfo = cpu_to_le32(si->nfinfo);
 	nilfs.segsum->ss_sumbytes = cpu_to_le32(si->sumbytes);
+	nilfs.segsum->ss_cno = cpu_to_le64(nilfs.cno);
 
 	/* initialize super root */
 	nilfs.super_root = map_disk_buffer(si->start + si->nblocks - 1, 1);
@@ -1559,8 +1600,7 @@ static void commit_segment(void)
 	segref->free_blocks_count = count_free_blocks(di);
 }
 
-static void prepare_super_block(struct nilfs_disk_info *di,
-				struct mkfs_options *opts)
+static void prepare_super_block(struct nilfs_disk_info *di)
 {
 	blocknr_t blocknr = NILFS_SB_OFFSET_BYTES / blocksize;
 	unsigned long offset = NILFS_SB_OFFSET_BYTES % blocksize;
@@ -1584,8 +1624,7 @@ static void prepare_super_block(struct nilfs_disk_info *di,
 	raw_sb->s_dev_size = cpu_to_le64(di->dev_size);
 	raw_sb->s_first_data_block = cpu_to_le64(di->first_segment_block);
 	raw_sb->s_blocks_per_segment = cpu_to_le32(di->blocks_per_segment);
-	raw_sb->s_r_segments_percentage =
-		cpu_to_le32(opts->r_segments_percentage);
+	raw_sb->s_r_segments_percentage = cpu_to_le32(r_segments_percentage);
 
 	raw_sb->s_ctime = cpu_to_le64(di->ctime);
 	raw_sb->s_mtime = 0;
@@ -1606,9 +1645,15 @@ static void prepare_super_block(struct nilfs_disk_info *di,
 	raw_sb->s_segment_usage_size =
 		cpu_to_le16(sizeof(struct nilfs_segment_usage));
 
+	raw_sb->s_feature_compat =
+		cpu_to_le64(compat_array[NILFS_FEATURE_TYPE_COMPAT]);
+	raw_sb->s_feature_compat_ro =
+		cpu_to_le64(compat_array[NILFS_FEATURE_TYPE_COMPAT_RO]);
+	raw_sb->s_feature_incompat =
+		cpu_to_le64(compat_array[NILFS_FEATURE_TYPE_INCOMPAT]);
+
 	uuid_generate(raw_sb->s_uuid);	/* set uuid using libuuid */
-	memcpy(raw_sb->s_volume_name, opts->volume_label,
-	       sizeof(opts->volume_label));
+	memcpy(raw_sb->s_volume_name, volume_label, sizeof(volume_label));
 }
 
 static void commit_super_block(struct nilfs_disk_info *di, 
