@@ -29,21 +29,54 @@
  *  - del_loop()
  */
 
-#define _LARGEFILE64_SOURCE
-#define _XOPEN_SOURCE 600
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif	/* HAVE_CONFIG_H */
 
+#if HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif	/* HAVE_SYS_TYPES_H */
+
 #include <stdio.h>
+
+#if HAVE_STDLIB_H
 #include <stdlib.h>
+#endif	/* HAVE_STDLIB_H */
+
+#if HAVE_UNISTD_H
 #include <unistd.h>
+#endif	/* HAVE_UNISTD_H */
+
+#if HAVE_FCNTL_H
 #include <fcntl.h>
+#endif	/* HAVE_FCNTL_H */
+
+#if HAVE_STRINGS_H
 #include <strings.h>
-#include <stdarg.h>
+#endif	/* HAVE_STRINGS_H */
+
+#if HAVE_STRING_H
 #include <string.h>
+#endif	/* HAVE_STRING_H */
+
+#if HAVE_SYS_MOUNT_H
 #include <sys/mount.h>
+#endif	/* HAVE_SYS_MOUNT_H */
+
+#if HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif	/* HAVE_SYS_STAT_H */
+
+#if HAVE_SYS_WAIT_H
 #include <sys/wait.h>
+#endif	/* HAVE_SYS_WAIT_H */
+
+#if HAVE_SYSLOG_H
+#include <syslog.h>
+#endif	/* HAVE_SYSLOG_H */
+
 #include <signal.h>
+#include <stdarg.h>
 #include <errno.h>
 
 #include "fstab.h"
@@ -52,12 +85,16 @@
 #include "xmalloc.h"
 #include "mntent.h"
 #include "mount_constants.h"
+#include "mount_opts.h"
 #include "mount.nilfs2.h"
+#include "cleaner_exec.h"
 #include "nls.h"
 
 
 int verbose = 0;
 int mount_quiet = 0;
+int readonly = 0;
+int readwrite = 0;
 static int nomtab = 0;
 
 extern char *optarg;
@@ -91,11 +128,29 @@ struct umount_options options = {
 /*
  * Other routines
  */
+static void nilfs_umount_logger(int priority, const char *fmt, ...)
+{
+	va_list args;
+
+	if ((verbose && priority > LOG_INFO) || priority >= LOG_INFO)
+		return;
+	va_start(args, fmt);
+	fprintf(stderr, "%s: ", progname);
+	vfprintf(stderr, fmt, args);
+	fputs(_("\n"), stderr);
+	va_end(args);
+}
+
+static void show_version(void)
+{
+	printf("%s (%s %s)\n", progname, PACKAGE, PACKAGE_VERSION);
+}
+
 static void parse_options(int argc, char *argv[], struct umount_options *opts)
 {
-	int c;
+	int c, show_version_only = 0;
 
-	while ((c = getopt(argc, argv, "nlfvr")) != EOF) {
+	while ((c = getopt(argc, argv, "nlfvrV")) != EOF) {
 		switch (c) {
 		case 'n':
 			nomtab++;
@@ -112,10 +167,18 @@ static void parse_options(int argc, char *argv[], struct umount_options *opts)
 		case 'r':
 			opts->remount++;
 			break;
+		case 'V':
+			show_version_only++;
+			break;
 		default:
 			break;
 		}
 	}
+
+	if (show_version_only) {
+		show_version();
+		exit(0);
+	}
 }
 
 static int umount_one(const char *, const char *, const char *, const char *,
@@ -154,14 +217,16 @@ int main(int argc, char *argv[])
 	struct umount_options *opts = &options;
 	int ret = 0;
 
-	parse_options(argc, argv, opts);
-
 	if (argc > 0) {
 		char *cp = strrchr(argv[0], '/');
 
 		progname = (cp ? cp + 1 : argv[0]);
 	}
 
+	nilfs_cleaner_logger = nilfs_umount_logger;
+
+	parse_options(argc, argv, opts);
+
 	umask(022);
 
 	if (opts->force)
@@ -328,8 +393,8 @@ umount_one(const char *spec, const char *node, const char *type,
 	if (mc) {
 		if (!read_only_mount_point(mc) &&
 		    (pid = get_mtab_gcpid(mc)) != 0) {
-			alive = check_cleanerd(spec, pid);
-			stop_cleanerd(spec, pid);
+			alive = nilfs_ping_cleanerd(pid);
+			nilfs_shutdown_cleanerd(spec, pid);
 		}
 	}
 
@@ -352,22 +417,23 @@ umount_one(const char *spec, const char *node, const char *type,
 				error(_("%s: could not remount %s read-only"),
 				      progname, spec);
 			}
-		} else if (alive && !check_cleanerd(spec, pid)) {
+		} else if (alive && !nilfs_ping_cleanerd(pid)) {
 			if (find_opt(mc->m.mnt_opts, pp_opt_fmt, &prot_period)
 			    < 0)
 				prot_period = ULONG_MAX;
 
-			if (start_cleanerd(spec, node, prot_period, &pid) == 0) {
+			if (nilfs_launch_cleanerd(spec, node, prot_period,
+						  &pid) == 0) {
 				if (verbose)
 					printf(_("%s: restarted %s(pid=%d)\n"),
-					       progname, CLEANERD_NAME,
+					       progname, NILFS_CLEANERD_NAME,
 					       (int)pid);
 				change_mtab_opt(spec, node, type,
 						change_gcpid_opt(opts, pid));
 				goto out;
 			} else
 				error(_("%s: failed to restart %s"),
-				      progname, CLEANERD_NAME);
+				      progname, NILFS_CLEANERD_NAME);
 		}
 	}
 
